import cv2
import numpy as np

def _DrawCell(bin_idx,bin_len=1,cell_size=(7,7)):
    h,w = cell_size
    cell = np.zeros((h,w))
    angle = np.pi*((bin_idx*20+90))/180
    h_len=  bin_len*h*np.sin(angle)
    w_len = bin_len*w*np.cos(angle)
    y1 = round((h-h_len)/2)
    y2 = round((h+h_len)/2)
    x1 = round((w+w_len)/2)
    x2 = round((w-w_len)/2)
    cv2.line(cell,(x1,y1),(x2,y2),1)
    return cell

def visualize_HOG(hist,cell_size=8):
    '''
    Parameters
    ----------
    - hist: HOG which returned from Cell_HOG or BlockNorm_HOG
    - cell_size: int

    Return
    ----------
    - hog_image : (M,N,C) ndarray

    '''
    h,w,c,b = hist.shape
    norm_hist = (255/np.max(hist))*hist
    canvas = None
    canvas = np.zeros((h*cell_size,w*cell_size,c))
    for i in range(h):
        for j in range(w):
            for k in range(b):
                for l in range(c):
                    canvas[i*cell_size:(i+1)*cell_size,j*cell_size:(j+1)*cell_size,l]+=norm_hist[i,j,l,k]*_DrawCell(k,norm_hist[i,j,l,k],(cell_size,cell_size))
    canvas = np.clip(canvas,0,255)         
    canvas = canvas.astype(np.uint8)
    if(c==1):
        canvas = np.reshape(canvas,(cell_size*h,cell_size*w))
    return canvas
    
def Fast_visualize_HOG(hist,cell_size=8):
    '''
    Parameters
    ----------
    - hist: HOG which returned from Cell_HOG or BlockNorm_HOG
    - cell_size: int

    Return
    ----------
    - hog_image : (M,N,C) ndarray

    '''
    h,w,c,b = hist.shape
    cell_ori = np.argmax(hist,axis=3)
    hist = np.max(hist,axis=3)
    norm_hist = (255/np.max(hist))*hist
    canvas = np.zeros((h*cell_size,w*cell_size,c))
    for i in range(h):
        for j in range(w):
            for l in range(c):
                canvas[i*cell_size:(i+1)*cell_size,j*cell_size:(j+1)*cell_size,l] = norm_hist[i,j,l]*_DrawCell(cell_ori[i,j,l],norm_hist[i,j,l],(cell_size,cell_size))
    canvas = np.clip(canvas,0,255)             
    canvas = canvas.astype(np.uint8)
    if(c==1):
        canvas = np.reshape(canvas,(cell_size*h,cell_size*w))
    return canvas




import cv2
import numpy as np

def _UnsignedOrientedGradients(img):
    dx = cv2.Sobel(img, cv2.CV_64F, 1, 0, ksize=1)
    dy = cv2.Sobel(img, cv2.CV_64F, 0, 1, ksize=1)
    mag, angle = cv2.cartToPolar(dx, -dy, angleInDegrees=True)

    angle = angle%180
    angle_map = None
    if(len(img.shape)==3):
        h,w,c = img.shape
        angle_map = np.zeros((h,w,3,9))
    else:
        h,w = img.shape
        angle_map = np.zeros((h,w,1,9))
        mag = mag[:,:,None]
        angle = angle[:,:,None]
    
    mask = angle>160
    angle_map[:,:,:,0] += (mag*mask)*(1-(180-mask*angle)/20)
    mask = angle<=20
    angle_map[:,:,:,0] += (mag*mask)*(1-mask*angle/20)
    for i in range(1,9):
        mask = np.bitwise_and(angle>(i-1)*20,angle<=(i+1)*20)
        angle_map[:,:,:,i] += (mag*mask)*(1-np.abs((i*20-mask*angle)/20))
    return angle_map

def _HOGinCell(angle_map,cell_size=(8,8)):
    h,w,c,_ = angle_map.shape
    h2 = int(h/cell_size[0])
    w2 = int(w/cell_size[1])
    cell_hist = np.zeros((h2,w2,c,9))
    tmp = np.zeros((h2,w,c,9))
    for i in range(h2):
        tmp[i] = np.sum(angle_map[i*cell_size[0]:(i+1)*cell_size[0]],axis=0)
    for i in range(w2):
        cell_hist[:,i] = np.sum(tmp[:,i*cell_size[1]:(i+1)*cell_size[1]],axis=1)
    return cell_hist

def Cell_HOG(img,cell_size=(8,8)):
    '''
    Parameters
    ----------
    - img: colorful or Gray scale (M,N,C) ndarry
    - cell_size: (int,int)

    Return
    ----------
    - cell_hist: (M/cell_size,N/cell_size,C,9) ndarry

    '''
    angle_map = _UnsignedOrientedGradients(img)
    cell_hist = _HOGinCell(angle_map,cell_size)
    return cell_hist

def BlockNorm_HOG(cell_hist,block_size=(2,2),flatten=False):
    '''
    Perform L2-norm block normalization

    Parameters
    ----------
    - cell_hist: HOG which returned from Cell_HOG or BlockNorm_HOG
    - block_size: (int,int)
    - flaten: boolean, ravel the out to a 1xn vector

    Return
    ------
    out: (n_blocks_row, n_blocks_col, channel, n_cells_row, n_cells_col, n_orient) ndarray
    '''
    eps=1e-5
    h,w,c,b = cell_hist.shape
    h2 = h-block_size[0]+1
    w2 = w-block_size[1]+1
    blockHOG = np.zeros((h2,w2,c,block_size[0],block_size[1],b))
    tmp = np.zeros((h,w2,c,block_size[1],b))
    for i in range(w2):
        for j in range(block_size[1]):
            tmp[:,i,:,j] = cell_hist[:,i+j]
    for i in range(h2):
        for j in range(block_size[0]):
            blockHOG[i,:,:,j] = tmp[i+j]
    blockHOG = blockHOG.reshape((h2,w2,c,-1))
    blocknormHOG = blockHOG/(np.sqrt(np.sum(blockHOG**2,axis=3)+eps**2)[:,:,:,None])
    blocknormHOG = blocknormHOG.reshape((h2,w2,c,block_size[0],block_size[1],b))
    if(flatten==False):      
        return blocknormHOG
    else:
        return np.ravel(blocknormHOG)

def HOGfeature(img,cell_size=(8,8),block_size=(2,2),flatten=False):
    '''
    Perform HOG in 9 orientations

    Parameters
    ----------
    - img: colorful or Gray scale (M,N,C) ndarry
    - cell_size: (int,int)
    - block_size: (int,int)
    - flaten: boolean, ravel the out to a 1xn vector

    Return
    ------
    out: (n_blocks_row, n_blocks_col, channel, n_cells_row, n_cells_col, n_orient) ndarray
    '''
    cell_hist = Cell_HOG(img,cell_size)
    blocknormHOG = BlockNorm_HOG(cell_hist,block_size,flatten)
    return blocknormHOG



import cv2
from hogBGR import HOG
from hogBGR import HOGvisualized

def NonBlockNormaliztionHOG_Show(img,cell_size=(8,8)):
    '''
    Parameters
    ----------
    - cell_hist: HOG which returned from Cell_HOG or BlockNorm_HOG
    - cell_size: int
    '''
    cell_hogfeature = HOG.Cell_HOG(img,cell_size)
    Img = HOGvisualized.visualize_HOG(cell_hogfeature,cell_size[0])
    cv2.imshow("No Block Normaliztion HOG",Img)
    cv2.waitKey(0)
    cv2.destroyWindow("No Block Normaliztion HOG")

def BlockNormaliztionHOG_Show(img,cell_size=(8,8),block_size=(2,2),one_cell=True):
    '''
    Parameters
    ----------
    - cell_hist: HOG which returned from Cell_HOG or BlockNorm_HOG
    - cell_size: int

    Return
    ----------
    - hog_image : (M,N,C) ndarray

    '''
    cell_hogfeature = HOG.Cell_HOG(img,cell_size)
    blocknormHOG = HOG.BlockNorm_HOG(cell_hogfeature,block_size)
    if(one_cell):
        Img = HOGvisualized.visualize_HOG(blocknormHOG[:,:,:,0,0,:],cell_size[0])
        cv2.imshow("HOG, first cell in each block",Img)
    else:
        Img=[]
        for i in range(block_size[0]):
            for j in range(block_size[1]):
                Img.append(HOGvisualized.visualize_HOG(blocknormHOG[:,:,:,i,j,:],cell_size[0]))
        for i in range(block_size[0]*block_size[0]):
            cv2.imshow("HOG, cell_{} in each block".format(i),Img[i])
    cv2.waitKey(0)
    cv2.destroyAllWindows()




import cv2
from hogBGR import HOG
from hogBGR import HOGvisualized
from hogBGR import util

def main():
    img = cv2.imread(r'hogBGR/example/rose.png')
    #img = cv2.imread(r'hogBGR/example/coins.png')
    cv2.imshow("Origin",img)
    '''
    util.BlockNormaliztionHOG_Show(img,one_cell=False)
    util.NonBlockNormaliztionHOG_Show(img)
    '''
    cell_hogfeature = HOG.Cell_HOG(img,cell_size=(16,16))
    #res = HOGvisualized.Fast_visualize_HOG(cell_hogfeature,7)
    res = HOGvisualized.visualize_HOG(cell_hogfeature,15)
    cv2.imshow("part1 - HOG feature without block normaliztion",res)

    hog = HOG.BlockNorm_HOG(cell_hogfeature,block_size=(2,2))
    res = HOGvisualized.visualize_HOG(hog[:,:,:,0,0,:],15)
    cv2.imshow("part2 - HOG feature",res)
    
    hog = HOG.HOGfeature(img,cell_size=(16,16),block_size=(2,2),flatten=False)
    res = HOGvisualized.visualize_HOG(hog[:,:,:,0,0,:],15)
    cv2.imshow("part3 - HOG feature (same result with part2)",res)

    cv2.waitKey(0)
    cv2.destroyAllWindows()

if __name__ == "__main__" :
    main()

    